try:
	import cv2
except:
	cv2 = None

def live_video(source=0,color_mode=None,resize=None,preprocess_function=None):
	"""
	Displays live stream on window. You can press `q` to stop.
	Parameters:
		source: Camera port default is 0
		color_mode: convert image to 'rgb' or 'grayscale'
		resize: (height , width) tuple resize incoming video frame
		preprocess_function: Submit preprocessing function if you need extra preprocessing for image
		Preprocessing Function Syntax:
		#defining function
		def preprocessing_function(frame): #RGB frame as input
			// do preprocessing stuff here
			return processed_frame (image as ndarray)

		#submitting to nv.live_video()
		nv.live_video(preprocessing_function)
	"""

	if cv2 is None:
		raise ImportError('OpenCV is required for running this function\n'
			'Please install cv2 using `pip install opencv-python`')

	video = cv2.VideoCapture(source)

	frame_height = int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))
	frame_width = int(video.get(cv2.CAP_PROP_FRAME_WIDTH))
	while True:
		# Capture frame-by-frame
		retval, frame = video.read()

		if resize is not None:
			if not isinstance(resize, tuple):
				raise ValueError(f'resize must be tuple of (height,width) but got {resize} instead')
			if (frame_height,frame_width) != resize:
				frame = cv2.resize(frame,resize)
		
		if color_mode is not None:
			if color_mode == 'rgb':
				frame = frame[...,::-1] #convert image to RGB
			elif color_mode == 'grayscale':
				frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY) #convert image to grayscale
			else:
				raise ValueError(f'Expected color_mode to be `rgb` or `grayscale` or None but got {color_mode} instead')
		if callable(preprocess_function): #check if preprocess_function is present
			frame = preprocess_function(frame)

		# Display the resulting frame
		cv2.imshow('frame', frame) #show image on screen

		if cv2.waitKey(1) & 0xFF == ord('q'):
			break

	# When everything is done, release the capture
	video.release()
	cv2.destroyAllWindows()


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from io import BytesIO
from warnings import warn
from re import sub
from base64 import b64encode,b64decode

try:
    from PIL import Image as pilimage
    from numpy import asarray as toarray
    from numpy import expand_dims as expand
    from numpy import squeeze
    from numpy import max as npmax
    from numpy import min as npmin
    import requests as request_image
except ImportError:
    request_image = None
    raise ImportError('Could not import PIL.Image. or Numpy '
        'This library requires PIL >= 7.0.0 and numpy >= 1.18.1')


interpolation_methods = {
        'nearest': pilimage.NEAREST,
        'bilinear': pilimage.BILINEAR,
        'bicubic': pilimage.BICUBIC,
        'hamming': pilimage.HAMMING,
        'box': pilimage.BOX,
        'lanczos': pilimage.LANCZOS
}


def imread(image_path,resize=None,color_mode = 'rgb',interpolation='nearest',dtype='float32',return_original = False):
    """Converts a PIL Image instance to a Ndarray optimised for model.
    Parameters
    ----------
        image_path: Image Path.
        resize: (width,height) tuple
        color_mode: default is `rgb`
            you can also use color_mode as`rgba` or `grayscale`
        interpolation:
            Interpolation method used to resample the image if the
            target size is different from that of the loaded image.
            Supported methods are "nearest", "bilinear", and "bicubic".
            If PIL version 1.1.3 or newer is installed, "lanczos" is also
            supported. If PIL version 3.4.0 or newer is installed, "box" and
            "hamming" are also supported.
            Default: "nearest".
        dtype: Dtype to use for the returned array.

        return_original: Returns original image array along with resized image array.
        	Default: False
        	Note: This parameter only works with resize parameter
    # Returns
        A 3D Numpy array.
    # Raises
        ValueError: if invalid `image_path` or `resize` or `color_mode` or `interpolation` or `dtype` is passed.
        ValueError: if return_original is True and resize is None
    """
    image = pilimage.open(image_path)
    if color_mode == 'grayscale':
        if image.mode not in ('L', 'I;16', 'I'):
            image = image.convert('L')
    elif color_mode == 'rgba':
        if image.mode != 'RGBA':
            image = image.convert('RGBA')
    elif color_mode == 'rgb':
        if image.mode != 'RGB':
            image = image.convert('RGB')
    else:
        raise ValueError('color_mode must be "grayscale", "rgb", or "rgba"')
        
    if resize is not None:
        if not isinstance(resize,tuple):
            raise TypeError(f'resize must be tuple of (width,height) but got {resize} of type {type(resize)} instead')

        if len(resize) != 2:
            raise ValueError(f'Tuple with (width,height) required but got {resize} instead.')

        original_image_array = toarray(image,dtype=dtype)
        if image.size != resize:
            if interpolation not in interpolation_methods:
                raise ValueError(f'Invalid interpolation, currently supported interpolations:{interpolation_methods.keys()}')
            resample = interpolation_methods.get(interpolation)
            image = image.resize(resize, resample)
    
    image_array = toarray(image,dtype=dtype)
    
    if return_original:
        if resize is None:
            raise ValueError("return_original parameter only works with resize parameter")
        return original_image_array , image_array
    
    return image_array


def imurl(image_url, return_as_array = False , **kwargs):
    """
    Read image from url and convert to bytes or ndarray
    
    Paramters
    ---------
    image_url: http / https url of image
    
    return_as_array: Convert image directly to numpy array
    	default: False
    
    kwargs:
        Keyword arguments of imread can be passed for image modification:
        Example:
            imurl(image_url,to_array=True,resize=(224,224),color_mode = 'rgb',dtype='float32')
            
        Note: kwargs only works with return_as_array = True
        
    Returns:
        PIL Image by default:
        if return_as_array is True:
            image will be returned as numpy array.
        
        Additional params like resize, color_mode, dtype , return_original can also be passed inorder to refine the image
    """
    if request_image is None:
        raise ImportError('requests library is required from reading image from url '
                         'Install it using pip install requests')
    
    
    if not image_url.startswith('http'):
        raise ValueError(f'invalid url found. Required http or https url but got {image_url} instead')
        
        
    image_response = request_image.get(image_url)
    imbytes = BytesIO(image_response.content)
    
    if return_as_array:
        return imread(imbytes,**kwargs)
    
    image = pilimage.open(imbytes)
    
    return image


def expand_dims(array,axis=0,normalize=False):
    """Expand the shape of an array.

    Insert a new axis that will appear at the `axis` position in the expanded
    array shape.

    Parameters
    ----------
    array : numpy array.
    axis : int or tuple of ints
    Position in the expanded axes where the new axis is placed
    normalize: 
        True : return normalized image
        False : return just image with expanded dimensions 
    """
    array = expand(array,axis=axis)
    if normalize:
        array /= 255.
    return array


def get_image_from_array(img_array, denormalize=True, dtype='float32' ,is_cv2_image = False):
    """Converts numpy image array to a PIL Image instance.
    Parameters
    ----------
        img: Input Numpy image array.
        denormalize: Revert back normalized image to unnormalized form
            Default: True.
        dtype: Dtype to use.
            Default: "float32".
        is_cv2_image: Set to True if image is loaded using cv2
            Default: False
    Returns
    -------
        A PIL Image.
    Raises
    ------
        Raises TypeError if image_array is not an numpy ndarray
    """
    if not hasattr(img_array, 'ndim'):
        raise TypeError(f'Required image_array to be of type numpy.ndarray but got {type(img_array)} instead')

    if img_array.ndim != 3:
        if img_array.ndim == 2:
            """expand image dimensions only if image is 2D grayscale
            manually adding channel dimension `1` to image (only for 2D grayscale image)"""
            img_array = expand_dims(img_array,axis=2)
        else:
            raise ValueError(f'Expected array with 3 dimensions Got array with shape {img_array.shape}\n'
                'Incase you have used expand_dims for preprocessing, use nv.reduce_dims() for reducing expanded dimensions\n'
                'make sure to check the axis position while expanding or reducing dimensions.')

    if is_cv2_image: #If numpy array is cv2 image
        img_array = img_array[...,::-1] #Convert BGR to RGB

    img_array = toarray(img_array,dtype=dtype)
    # Original Numpy array x has format (height, width, channel)
    # or (channel, height, width)
    # but target PIL image has format (width, height, channel)

    if denormalize:
        img_array = img_array - npmin(img_array)
        img_max = npmax(img_array)
        if img_max != 0:
            img_array /= img_max
        img_array *= 255
    if img_array.shape[2] == 4: #RGBA Image
        return pilimage.fromarray(img_array.astype('uint8'), 'RGBA')
    elif img_array.shape[2] == 3: #RGB image
        return pilimage.fromarray(img_array.astype('uint8'), 'RGB')
    elif img_array.shape[2] == 1: # grayscale image
        if npmax(img_array) > 255:
            # 32-bit signed integer grayscale image. PIL mode "I"
            return pilimage.fromarray(img_array[:, :, 0].astype('int32'), 'I')
        return pilimage.fromarray(img_array[:, :, 0].astype('uint8'), 'L')
    else:
        raise ValueError(f'Channel {img_array.shape[2]} not supported')


def imshow(image,is_cv2_image=False):
    """
    Displays image in new window
    Parameters
    image: PIL or CV2 image array
    is_cv2_image: If image_array is processed using cv2
    """
    if hasattr(image,'show'):
        image.show()
    else:
        get_image_from_array(image,is_cv2_image=is_cv2_image).show()


def imsave(path,image,file_format = None ,is_cv2_image=False,denormalize=True,**kwargs):
    """
    Write image array or instance to a file.
    Parameters
    ----------
    path: Location for writing image file
    image: image array
    is_cv2_image: Set to True if image is loaded using cv2 
        Default: False
    denormalize: Set to True if image was normalized during preprocessing
        Default: True
    """
    if hasattr(image,'save'): 
        image.save(path,file_format=file_format,**kwargs)
    else:
        image = get_image_from_array(image, denormalize=denormalize , is_cv2_image=is_cv2_image)
        if image.mode == 'RGBA' and (file_format == 'jpg' or file_format == 'jpeg'):
            warn('JPG format does not support RGBA images, converting to RGB.')
            image = image.convert('RGB')
        image.save(path, format=file_format, **kwargs)


def reduce_dims(array,axis=0):
    """
    Reduce array dimensions at given axis
    Note: If trying on image array please check if expand_dims 
          is used by you during image preprocessing and axis too.

    Parameters
    ----------
    array: numpy nd array
    axis : int or tuple of ints
    Position in the expanded axes where the new axis is placed.
    default: 0
    """
    return squeeze(array,axis=axis)


def base64_to_bytes(base64_encoded_image):
    """
    Convert base64 image data to PIL image

    Parameters
    ----------
    base64_encoded_image: base64 encoded image

    Returns
    -------
    Decoded image as Bytes Array
    """
    image_data = sub('^data:image/.+;base64,', '', base64_encoded_image)
    return BytesIO(b64decode(image_data))


def image_to_base64(image_from_array,file_format='PNG'):
    """
    Convert image from array to base64 string
    parameters
    ----------
    image_from_array: PIL image instance 
    	Incase image is numpy array, convert to image object using nv.get_image_from_array(array)

    returns
    -------
    base64 encoded image as string
    """
    buffered = BytesIO()
    image_from_array.save(buffered, format=file_format)
    return u"data:image/png;base64," + b64encode(buffered.getvalue()).decode("ascii")


from ._version import version as __version__
from .image_utils import imread, imurl, expand_dims , reduce_dims , imshow , imsave , get_image_from_array , base64_to_bytes , image_to_base64 
from .video_utils import live_video



from setuptools import setup
from nsvision import __version__
with open("README.md", "r") as fh:
    long_description = fh.read()

classifiers = [
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "License :: OSI Approved :: MIT License",
    "Topic :: Utilities",
]

setup(
    name="nsvision",
    version=__version__,
    python_requires=">=3.6",
    description="nsvision - Computer Vision Wrapper built on top of PIL, cv2 and Numpy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Nsemble/nsvision",
    author="Nsemble.ai",
    author_email="admin@nsemble.ai",
    license="MIT",
    packages=["nsvision"],
    install_requires=[
        'numpy==1.18.2',
        'Pillow==7.0.0'
    ],
    classifiers=classifiers,
    scripts=[
    "bin/split_data",
    "bin/rename_files",
    "bin/tumor_data_extractor",
    "bin/split_data_gui"
    ],
)



import os 
from shutil import copy2
from random import sample
from numpy import array as nparray
from nsvision import get_image_from_array
try:
	from natsort import natsorted
	from h5py import File as h5py_file
except:
	natsorted = None
	h5py_file = None


def check_ratio_sum(ratio):
	"""
	Check the given ratio is valid or not. Sum of ratio should be 100
	"""

	train_ratio = int(ratio[0])
	val_ratio = int(ratio[1])
	test_ratio = int(ratio[2])
	qa_ratio = int(ratio[3])
	total = train_ratio + test_ratio + val_ratio + qa_ratio
	if total != 100:
		raise ValueError("Sum of all the input ratio should be equal to 100 ")
	return train_ratio, val_ratio, test_ratio, qa_ratio


def image_frequency(num,ratio):
	"""
	Find the frequency per class
	Parameter
	---------
	num: Total number of images
	ratio: split ratio
	"""
	ratio = ratio/100
	return round( num * ratio, 0 )


def split_image_data(data_dir,ratio,generate_labels_txt=False):
	"""
	Divides image data required for classification according to Train, Test, Validation and QA

	Parameters:
	data_dir:  path to image data folder
	ratio: Tuple ratio
	generate_labels_txt: set to True if you want to generate labels.txt for the classes.
		Default(False)
	"""
	if not isinstance(ratio, tuple):
		raise TypeError(f"ratio should be tuple of (train,val,test,qa) but got {ratio} of type {type(ratio)} instead")

	if len(ratio) != 4:
		raise ValueError(f"required ratio to be of length 4 but got length {len(ratio)} instead")
	#ratio 
	train_ratio, val_ratio, test_ratio, qa_ratio = check_ratio_sum(ratio)

	#making new_dir to save divided data
	folder_name = os.path.splitext(os.path.basename(data_dir))[0]
	new_folder = folder_name + "_classified_data"
	model_data_dir = os.path.join(os.path.dirname(data_dir),new_folder)
	os.makedirs(model_data_dir,exist_ok = True)

	os.makedirs(os.path.join(model_data_dir,"train"),exist_ok = True)
	os.makedirs(os.path.join(model_data_dir,"test"),exist_ok = True)
	os.makedirs(os.path.join(model_data_dir,"val"),exist_ok = True)
	os.makedirs(os.path.join(model_data_dir,"qa"),exist_ok = True)

	folder_names = ['train','val','test','qa']

	#dividing the data into given ratio
	classes_list = os.listdir(data_dir)

	print("Total number of classes",len(classes_list))
	try:
		for class_name in classes_list:
			class_path = os.path.join(data_dir,class_name)
			
			class_images = os.listdir(class_path)
			
			total_class_images = len(class_images)

			total_train_images = image_frequency(total_class_images,train_ratio)
			total_val_images = image_frequency(total_class_images,val_ratio)
			total_test_images = image_frequency(total_class_images,test_ratio)
			total_qa_images = image_frequency(total_class_images,qa_ratio)

			ratio_list = [total_train_images,total_val_images,total_test_images,total_qa_images]

			class_images_new = class_images
			class_images_new = sample(class_images_new,len(class_images_new))

			div1 = total_train_images
			div2 = div1 + total_val_images
			div3 = div2 + total_test_images
			print(f"\nData divided for class {class_name} as follows:-",f"Train: {div1}",f"Validation: {div2 - div1}", f"Test: {div3 -div2}",f"QA: {total_class_images - div3}\n",sep='\n')
			train_list = class_images_new[:int(div1)]
			val_list = class_images_new[int(div1):int(div2)]
			test_list = class_images_new[int(div2):int(div3)]
			qa_list = class_images_new[int(div3):]
			split_lst = [train_list,val_list,test_list,qa_list]
			for i,j in zip(folder_names,split_lst):
				for img in j:
					img_path = os.path.join(class_path,img)
					if os.path.isfile(img_path):
						# src_path = os.path.join(class_path,img)
						dst_path = os.path.join(os.path.join(model_data_dir,i),class_name)
						os.makedirs(dst_path,exist_ok=True)
						copy2(img_path,dst_path)
		if generate_labels_txt:
			print("Generating labels.txt")
			with open(os.path.join(os.path.dirname(data_dir),'labels.txt'),'w') as labels_file:
				labels_file.write("\n".join(classes_list))
	except:
		raise Exception("Failed to split data please check folder structure or check your OS Permission")
	print("Splitting completed",f"Splitted data is stored at {model_data_dir}",sep='\n')


def rename_files(name,folder_path,number):
	"""
	Rename the files in the given folder path
	Parameter
	---------
	path - folder path
	name - common name for renaming
	number - number from which renaming is to start
	"""
	if natsorted is None:
		raise ImportError("natsorted is required"
			"Install it using `pip install natsort`")


	folder_name = os.path.splitext(os.path.basename(folder_path))[0]
	new_folder = folder_name + "_renamed"
	dst_path = os.path.join(os.path.dirname(folder_path),new_folder)
	os.makedirs(dst_path,exist_ok = True)
	for i,filename in enumerate(natsorted(os.listdir(folder_path)),number):
		old_path = os.path.join(folder_path,filename)
		try:
			extension = os.path.splitext(os.path.basename(old_path))[1]
			if extension != '':
				new_name = f"{name}_{i}{extension}"
				new_path = os.path.join(dst_path,new_name)
				copy2(old_path,new_path)
		except:
			raise Exception("Failed to rename the files in the given folder please check the folder structure or os permission")
	print("Renaming completed",f"Renamed files are stored at {dst_path}",sep='\n')


def get_image_from_mat(mat_filepath):
	"""
	This function convert .mat file to pil image
	Arguments  -
	Input - 1. matfile path (type - str)
			2. data (type - boolean)

	Output - pil_image, cjdata
	"""
	if h5py_file is None:
		raise ImportError("h5py is required"
			"Install using `pip install h5py`")


	mat_file =  h5py_file(mat_filepath,'r')
	cjdata = mat_file['cjdata']
	image_array = nparray(cjdata.get('image')).astype("float64")
	nv_image = get_image_from_array(image_array)
	label = cjdata.get('label')[0][0]
	return nv_image, label
